;;  Connected components in the grid -- for simulator   File:  grcc.lisp

(in-package '*lisp)

(*COLD-BOOT :initial-dimensions '(8 8))

(*DEFVAR CC (!! 0))    ; Component number
(*DEFVAR CW (!! 0))    ; Working component
(*DEFVAR X  (!! 0))    ; X address
(*DEFVAR Y  (!! 0))    ; Y address

(defvar xmx 7) (defvar ymx 7)   ; Maximum grid addresses

(*DEFVAR NB NIL!!)     ; Barrier on North, etc.
(*DEFVAR EB NIL!!)
(*DEFVAR WB NIL!!)
(*DEFVAR SB NIL!!)

(defun setup (p)
"Setup, 0 <= p <= 100 (integer) is probability of barrier"
 (*set CC (!! 0))
 (*set X (self-address-grid!! (!! 0)))  (*set Y (self-address-grid!! (!! 1)))
 (setq xmx (1- (dimension-size 0)))     (setq ymx (1- (dimension-size 1)))
 (*set NB (zerop!! X))  (*set SB (=!! X (!! xmx)))
 (*set WB (zerop!! Y))  (*set EB (=!! Y (!! ymx)))
 (*when (and!! (<!! X (!! xmx)) (<!! (random!! (!! 100)) (!! p)))
    (*set SB T!!)  (*news T!! NB 1 0))
 (*when (and!! (<!! Y (!! ymx)) (<!! (random!! (!! 100)) (!! p)))
    (*set EB T!!)  (*news T!! WB 0 1))
)

(defun gccs ()
"Calculate connected components, returns no. of cycles"
 (*set CC (self-address!!))
 (do ((n 0 (incf n)))
   ((not (gccx)) n))
)

(defun gccx ()
"Perform one cycle of sweeps, test for changes"
 (let (r)
  (*set CW CC)
  (*set CW
   (scan!! CW 'min!! :dimension 1 :direction :forward  :segment-pvar WB))
  (*set CW
   (scan!! CW 'min!! :dimension 1 :direction :backward :segment-pvar EB))
  (*set CW
   (scan!! CW 'min!! :dimension 0 :direction :forward  :segment-pvar NB))
  (*set CW
   (scan!! CW 'min!! :dimension 0 :direction :backward :segment-pvar SB))
  (setq r (*or (not!! (=!! CC CW))))
  (*set CC CW)
  r)
)

(defun gccy ()
"Perform one cycle of sweeps, test for changes.  Uses copy on rev scan.
 Does not update CC."
 (*set CW CC)
 (*set CW
  (scan!! CW 'min!!  :dimension 1 :direction :forward  :segment-pvar WB))
 (*set CW
  (scan!! CW 'copy!! :dimension 1 :direction :backward :segment-pvar EB))
 (*set CW
  (scan!! CW 'min!!  :dimension 0 :direction :forward  :segment-pvar NB))
 (*set CW
  (scan!! CW 'copy!! :dimension 0 :direction :backward :segment-pvar SB))
 (*or (/=!! CC CW))
)

(*DEFVAR RC NIL!!)             ;; Subregion center
(*DEFVAR SA (self-address!!))  ;; Self-address, for convenience
(*DEFVAR CY (!! 0))            ;; Working comp. no.

(defun updc ()
"Update centers, CW should have new comp. nos., CW <= CC."
  (*when RC
    (*set CY CW)
    (*when (<!! CW CC SA) (*pset :min CW CY CC)))
    (*set CW (min!! CW CY))
    (*when (<!! CW CC) (*set CW (pref!! CW CW)))
)

(defun gcct ()
"Calculate connected components using centers"
 (prog ((n 1))
       (*set CC SA)  (*set RC NIL!!)
       (gccy)  (*set CC CW)
       (*pset :or T!! RC CC)
  loop (if (not (gccy)) (return n))
       (updc)  (*set CC CW)  (incf n)
       (go loop))
)

(*defun bline (B i)
"Draw line according to row i of B"
 (princ "+")
 (dotimes (j (1+ ymx)) (princ (if (pref B (grid i j)) "--+" "  +")))
 (terpri)
)

(defun dcc()
"Display current grid and component status"
 (bline NB 0)
 (dotimes (i (1+ xmx))
   (princ "|")
   (dotimes (j (1+ ymx))
     (format t "~2D" (pref CC (grid i j)))
     (princ (if (pref EB (grid i j)) "|" " ")))
   (terpri)
   (bline SB i))
)